local cv_enabled = CV_RegisterVar({
    name = "luaspeedpads",
    defaultvalue = "On",
    possiblevalue = CV_OnOff,
    flags = CV_NETVAR,
})

local cv_warnings = CV_RegisterVar({
	name = "luaspeedpads_warnings",
	defaultvalue = "Off",
	possiblevalue = CV_OnOff,
})

local speedpadparam = {} -- [tag] = line_index
local speedpadparamwarn = {} -- [tag] = bool -- To only warn about params missing once per tag on map
local sectorcentercache = {} -- [sector_index] = vec2

addHook("MapLoad", function()
    -- Always run this, so we can hot-switch luaspeedpads On and Off

    speedpadparam = {}
    speedpadparamwarn = {}
    sectorcentercache = {}

    for line in lines.iterate do
        if line.special == 4 and speedpadparam[line.tag] == nil then
            speedpadparam[line.tag] = #line
        end
    end
end)

addHook("NetVars", function(net)
    speedpadparam = net($)
    sectorcentercache = net($) -- Probably not necessary
end)

local function warn(...)
	if not cv_warnings.value then return end
	print(...)
end

local function getSpeedPadParams(tag)
    local i = speedpadparam[tag]

    if i == nil then return end

    return lines[i]
end

local function getSectorCenter(sector)
    local center = sectorcentercache[#sector]

    if center == nil then
        -- Find bounding box
        local minx, maxx
        local miny, maxy

        for i = 0, #sector.lines-1 do
            local line = sector.lines[i]

            if minx == nil then
                minx = min(line.v1.x, line.v2.x)
                maxx = max(line.v1.x, line.v2.x)

                miny = min(line.v1.y, line.v2.y)
                maxy = max(line.v1.y, line.v2.y)
            else
                minx = min($, min(line.v1.x, line.v2.x))
                maxx = max($, max(line.v1.x, line.v2.x))

                miny = min($, min(line.v1.y, line.v2.y))
                maxy = max($, max(line.v1.y, line.v2.y))
            end
        end

        -- Center of sector would be center of bounding box
        -- doing >>FRACBITS before addition to avoid int32 overflows
        center = { x = ((minx/FRACUNIT + maxx/FRACUNIT)/2)*FRACUNIT, y = ((miny/FRACUNIT + maxy/FRACUNIT)/2)*FRACUNIT }

        sectorcentercache[#sector] = center
    end

    return center
end

local cv_kartvoices
local function useVoices()
    if not cv_kartvoices then
        cv_kartvoices = CV_FindVar("kartvoices") or { value = 1 }
    end

    return cv_kartvoices.value
end

local function P_MobjSlopeZ(slope, mo, invert, z)
    if not slope then return z end

    local testx, testy

    if slope.d.x < 0 then
        testx = mo.radius
    else
        testx = -mo.radius
    end

    if slope.d.y < 0 then
        testy = mo.radius
    else
        testy = -mo.radius
    end

    if (slope.zdelta > 0) == invert then
        testx = -testx
        testy = -testy
    end

    testx = testx + mo.x
    testy = testy + mo.y

    return P_GetZAt(slope, testx, testy)
end

local function P_MobjFOFTopZ(ffloor, mo)
    return P_MobjSlopeZ(ffloor.t_slope, mo, true, ffloor.topheight)
end

local function P_MobjFOFBottomZ(ffloor, mo)
    return P_MobjSlopeZ(ffloor.b_slope, mo, false, ffloor.bottomheight)
end

local function P_MobjFloorZ(sector, mo)
    return P_MobjSlopeZ(sector.f_slope, mo, true, sector.floorheight)
end

local function P_MobjCeilingZ(sector, mo)
    return P_MobjSlopeZ(sector.c_slope, mo, false, sector.ceilingheight)
end

-- For regular sectors only
local function collidesWithSector(mo, sector)
    local top = P_MobjCeilingZ(sector, mo)
    local bottom = P_MobjFloorZ(sector, mo)

    local tag = sector.tag

    return (mo.z <= bottom and SG_FloorIsTrigger(tag)) or ((mo.z + mo.height) >= top and SG_CeilingIsTrigger(tag))
end

-- For FOF's only
local function collidesWithFOF(mo, ffloor)
    if not (ffloor.flags & FF_EXISTS) then return end

    local top = P_MobjFOFTopZ(ffloor, mo)
    local bottom = P_MobjFOFBottomZ(ffloor, mo)

    -- Dunno how but it happens
    if bottom > top then
        top, bottom = bottom, top -- uwu
    end

    --print("top="..top..", bottom="..bottom..", mo.z="..mo.z..", mo.z+height="..(mo.z+mo.height))

    return mo.z <= top and (mo.z + mo.height) >= bottom
end

-- Do the speedpad stuff
-- Port of hardcode version, for most part
local function doSpeedPad(p, sector, ffloor)
    local pmo = p.mo
    local params = getSpeedPadParams(sector.tag)

    if not params then
        if not speedpadparamwarn[sector.tag] then
            warn("\x82WARNING:\x80 speed pad with no params found for tag "..sector.tag)
            speedpadparamwarn[sector.tag] = true
        end

        return
    end

    local lineangle = R_PointToAngle2(params.v1.x, params.v1.y, params.v2.x, params.v2.y)

    -- Always update the timer for forced angle
    p.dashpadangletimer = TICRATE/3

    if p.dashpadcooldown > 0 then
        return
    end

    -- But only reset angle itself when dashpad activates
    p.dashpadangle = lineangle

    local linespeed = FixedHypot(params.v2.x-params.v1.x, params.v2.y-params.v1.y)

    pmo.angle = lineangle

    -- SRB2Kart: Scale the speed you get from them!
    -- This is scaled differently from other horizontal speed boosts from stuff like springs, because of how this is used for some ramp jumps.
    if pmo.scale > mapobjectscale then
        -- FixedDiv is there to get rid of mapobjectscale multiplier. If you have pmo.scale = 2*mapobjectscale, you get 2x boost
        --linespeed = FixedMul(linespeed, FixedDiv(pmo.scale, mapobjectscale))
        linespeed = FixedMul(linespeed, pmo.scale) -- This is vanilla-compatible tho
    end

    -- localangle assignment used to be there - not accessible in lua soooo

    -- Teleport to sector center
    if not (params.flags & ML_EFFECT4) then
        local center = getSectorCenter(ffloor and ffloor.target or sector)

        -- Why do this weird substraction? Because vanilla speedpads activate 1 frame later, and a certain "pixel-perfect"
        -- zipper doesn't work as intended
        --P_MoveOrigin(pmo, center.x - FixedMul(linespeed, cos(lineangle)), center.y - FixedMul(linespeed, sin(lineangle)), pmo.z) -- This fixes volcanic crater zipper...
        P_MoveOrigin(pmo, center.x, center.y, pmo.z) -- But this doesn't let you stuck in ballistic reef if you hold brake
    end

    P_InstaThrust(pmo, pmo.angle, linespeed)

    p.dashpadcooldown = TICRATE/3

    -- No :Trollic:
    --p.kartstuff[k_drift] = 0
    --p.kartstuff[k_driftcharge] = 0

    p.kartstuff[k_pogospring] = 0

    S_StartSound(pmo, sfx_spdpad)
    --S_StartSound(pmo, sfx_cdfm62) -- Different sound mostly just to distinguish lua version from vanilla

    local pick = P_RandomKey(2) -- Gotta roll the RNG every time this is called for sync reasons
    if useVoices() then
        S_StartSound(pmo, sfx_kbost1+pick)
    end
    --K_TauntVoiceTimers(player)
end

addHook("PlayerSpawn", function(p)
    p.dashpadcooldown = 0
end)

addHook("PlayerThink", function(p)
    if cv_enabled.value == 0 then return end

    -- This makes vanilla zipper panels not activate
    p.kartstuff[k_dashpadcooldown] = 1
end)

local function detectSpeedpads(p)
    if not p.mo then return end

    local pmo = p.mo

    p.dashpadcooldown = max(($ or 0) - 1, 0)

    -- Check sector we're in first
    local sec = pmo.subsector.sector
    local spec = GetSecSpecial(sec.special, 3)

    if (spec == 5 or spec == 6) and collidesWithSector(pmo, sec) then
        doSpeedPad(p, sec, nil)
        -- If speedpads are overlapping, it will skip them. Dunno if thats vanilla behavior but i never saw that happen on maps anyway soooo
        return
    end

    for ffloor in sec.ffloors() do
        local control = ffloor.sector

        spec = GetSecSpecial(control.special, 3)

        if (spec == 5 or spec == 6) and collidesWithFOF(pmo, ffloor) then
            doSpeedPad(p, control, ffloor)
            return
        end
    end
end

addHook("PlayerThink", function(p)
    if cv_enabled.value == 0 then return end

    local pmo = p.mo

    if not pmo then return end

    if not P_IsObjectOnGround(pmo) then
        p.dashpadangletimer = 0
        return
    end

    p.dashpadangletimer = max(($ or 0) - 1, 0)

    if p.dashpadangletimer then
        P_InstaThrust(pmo, p.dashpadangle, FixedHypot(pmo.momx, pmo.momy))
    end
end)

-- This makes K_KartPlayerThink not spawn postimages (if we don't need to)
addHook("PreThinkFrame", function()
    if cv_enabled.value == 0 then return end

    for p in players.iterate do
        p.kartstuff[k_dashpadcooldown] = p.dashpadcooldown or 0
    end
end)

addHook("PostThinkFrame", function()
    if cv_enabled.value == 0 then return end

    for p in players.iterate do
        detectSpeedpads(p)
    end
end)
