-- Allows creating different appearances for same character (for example,
-- different sonic sprites) without taking skin slot
--
-- Idea by legion

local appearance_version = 1

if appearance_loaded then
    if appearance_loaded < appearance_version then
        error("Older version of appearance loaded, this may lead to problems")
    end

    print("\131NOTICE:\128 Appearance already loaded (version #"..appearance_loaded..")")
    return
end

rawset(_G, "appearance_loaded", appearance_version)

local appearance = {}
-- G: signs had some mind-boggling bugs when reading pmo.sprite
rawset(_G, "APPEAR_Skins", appearance)

-- Hud gfx
local appearance_hud = {}
rawset(_G, "APPEAR_HUD", appearance_hud)

-- CONFIG
local config = "appearance.cfg"

local function loadConfig()
    local file = io.open(config, "r")

    -- No config :(
    if not file then
        return
    end

    for line in file:lines() do
        COM_BufAddText(consoleplayer, line)
    end

    file:close()
end

local function saveConfig()
    if not consoleplayer then return end

    local file = io.open(config, "w")

    if not file then return end

    if consoleplayer.appearance ~= nil then
        for skin, choice in pairs(consoleplayer.appearance) do
            file:write("appear "..skin.." "..choice..' true\n')
        end
    end

    file:close()
end

addHook("PlayerJoin", function(pnum)
    if consoleplayer and #consoleplayer == pnum then
        loadConfig()
    end
end)

-- CONFIG

rawset(_G, "APPEAR_AddAppearance", function(skin, appearname, spritename, rank, want)
    if appearname == "default" then error("appearance name 'default' is reserved") end

    skin = skin:lower()
    appearname = appearname:lower()
    spritename = "SPR_"..spritename:upper()

    freeslot(spritename)

    if not appearance[skin] then appearance[skin] = {} end

    appearance[skin][appearname] = _G[spritename]

    if not appearance_hud[skin] then appearance_hud[skin] = {} end

	if rank then
		appearance_hud[skin][appearname] = { facerank = rank, facewant = want }
	end
end)

-- Returns player's appearance table or creates new if necessary
rawset(_G, "APPEAR_GetAppearanceTable", function(player)
    if not player.appearance then player.appearance = {} end

    return player.appearance
end)

-- Returns currently used appearance for player. If player doesn't have mobj and no skin specified, it will return nil
rawset(_G, "APPEAR_GetAppearance", function(player, skin)
    if not (player.mo or skin) then return end

    if not skin then skin = player.mo.skin end

    if not player.appearance then return "default" end

    return player.appearance[skin] or "default"
end)

-- Returns name for rank gfx. If player doesn't have mobj, it will return nil
rawset(_G, "APPEAR_GetRankGFX", function(player, skin)
    if not (player.mo or skin) then return end

    skin = skin or player.mo.skin
    local default = skins[skin].facerank
    local appear = APPEAR_GetAppearanceTable(player)

    if appear[skin] then
        local appear_hud = appearance_hud[skin] and appearance_hud[skin][appear[skin]]

        if appear_hud then
            return appear_hud.facerank or default
        end
    end

    return default
end)

-- Returns name for wanted gfx. If player doesn't have mobj, it will return nil
rawset(_G, "APPEAR_GetWantedGFX", function(player, skin)
    if not (player.mo or skin) then return end

    skin = skin or player.mo.skin
    local default = skins[skin].facewant
    local appear = APPEAR_GetAppearanceTable(player)

    if appear[skin] then
        local appear_hud = appearance_hud[skin] and appearance_hud[skin][appear[skin]]

        if appear_hud then
            return appear_hud.facewant or default
        end
    end

    return default
end)

COM_AddCommand("appear", function(player, skin, appearname, config_loading)
    local appear = APPEAR_GetAppearanceTable(player)

    if skin then skin = skin:lower() end
    if appearname then appearname = appearname:lower() end

    if not skin then
        CONS_Printf(player, "Usage: appear\128 <\130skin\128> [\132name\128]")
        CONS_Printf(player, "Without appearance name, will print all available variants for it")
        return
    end

    if not appearname then
        if not appearance[skin] then
            CONS_Printf(player, "\134This skin doesn't have any custom appearances")
        else
            local buff = {"Available appearances for skin \130"..skin.."\128:", "default"}

            for name, _ in pairs(appearance[skin]) do
                table.insert(buff, name)
            end

            CONS_Printf(player, table.concat(buff, " \132"))
        end
        return
    end

    if appearname == "default" then
        appear[skin] = nil
        if player == consoleplayer and not config_loading then saveConfig() end
        CONS_Printf(player, "Set \132default\128 appearance for skin \130"..skin)
        return
    end

    if not appearance[skin] then
        CONS_Printf(player, "\134This skin doesn't have any custom appearances")
        return
    end

    if not appearance[skin][appearname] then
        CONS_Printf(player, "Appearance \132"..appearname.."\132 \133does not exist\128 for this skin")
        return
    end

    appear[skin] = appearname
    CONS_Printf(player, "Set appearance \132"..appearname.."\128 for skin \130"..skin)
    if player == consoleplayer and not config_loading then saveConfig() end
end)

COM_AddCommand("appear_list", function(player)
    local appears = {"Skins with custom appearances:"}

    for skin, _ in pairs(appearance) do
        local skin_appears = {}

        for appearname, _ in pairs(appearance[skin]) do
            table.insert(skin_appears, appearname)
        end

        table.insert(appears, "\130"..skin.."\128: \132"..table.concat(skin_appears, ", "))
    end

    if #appears == 1 then
        CONS_Printf(player, "\134There are no custom appearances yet...")
    else
        CONS_Printf(player, table.concat(appears, "\n"))
    end
end)

addHook("ThinkFrame", function()
    for player in players.iterate do
        if player.mo then
            local pmo = player.mo
            local appear = APPEAR_GetAppearanceTable(player)

            if appear[pmo.skin] then
                local sprite = appearance[pmo.skin][appear[pmo.skin]]
                pmo.sprite = sprite

				-- G: don't need this anymore
				--[[
                if player.signoverlay then
                    local delete_overlays = {}

                    -- pairs iterator is not synch-safe, but there aren't any
                    -- code that relies on execution order anyway
                    for overlay, _ in pairs(player.signoverlay) do
                        if overlay.valid then
                            overlay.sprite = sprite
                        else
                            -- afaik removing table entry while iterating over
                            -- it is not safe so imma do this instead
                            table.insert(delete_overlays, overlay)
                        end
                    end

                    -- Now we can safely remove everything
                    for _, overlay in ipairs(delete_overlays) do
                        player.signoverlay[overlay] = nil
                    end
                end
				--]]
            end
        end
    end
end)

-- G: or this
--[[
addHook("MobjThinker", function(mo)
    if not (mo.target and mo.target.valid) then return end

    local sign = mo.target

    if sign.target and sign.target.valid and sign.target.player then
        local pmo = sign.target
        local player = pmo.player

        if not player.signoverlay then player.signoverlay = {} end

        -- Using table as a set
        player.signoverlay[mo] = true
    end
end, MT_OVERLAY)
--]]
